import gzip
import typing as ta

from omlish import lang
from omlish.formats import json


##


@lang.cached_function()
def load_instance_types() -> ta.Mapping[str, ta.Mapping[str, ta.Any]]:
    raw = lang.get_relative_resources(globals=globals())['cache.json.gz'].read_bytes()
    data = gzip.decompress(raw)
    return json.loads(data)



import typing as ta

from omlish import cached
from omlish import check
from omlish import collections as col
from omlish import dataclasses as dc
from omlish import lang


DateTime = ta.NewType('DateTime', str)
MillisecondDateTime = ta.NewType('MillisecondDateTime', str)

Timestamp = ta.NewType('Timestamp', str)


##


@dc.dataclass(frozen=True)
class Tag:
    key: str
    value: str


TagList: ta.TypeAlias = ta.Sequence[Tag]


##


class ValueType(lang.Abstract):
    pass


@dc.dataclass(frozen=True)
class ListValueType(ValueType):
    e: type


@dc.dataclass(frozen=True)
class MapValueType(ValueType):
    k: type
    v: type


##


class SHAPE_NAME(lang.Marker):  # noqa
    pass


def common_metadata(
        *,
        shape_name: str | None = None,
) -> dict[ta.Any, ta.Any]:
    md = {}

    if shape_name is not None:
        md[SHAPE_NAME] = shape_name

    return md


##


def shape_metadata(
        **kwargs: ta.Any,
) -> dict[ta.Any, ta.Any]:
    md = {**common_metadata(**kwargs)}

    return md


class ShapeInfo:
    def __init__(
            self,
            cls: type['Shape'],
            metadata: ta.Mapping[ta.Any, ta.Any],
    ) -> None:
        super().__init__()

        self._cls = check.issubclass(cls, Shape)
        self._metadata = metadata

    @property
    def cls(self) -> type['Shape']:
        return self._cls

    @property
    def metadata(self) -> ta.Mapping[ta.Any, ta.Any]:
        return self._metadata

    #

    @cached.property
    def fields(self) -> ta.Sequence[dc.Field]:
        check.state(dc.is_immediate_dataclass(self._cls))
        fls = dc.fields(self._cls)
        return fls  # noqa

    @cached.property
    def fields_by_name(self) -> ta.Mapping[str, dc.Field]:
        return col.make_map_by(lambda fl: fl.name, self.fields, strict=True)

    @cached.property
    def fields_by_member_name(self) -> ta.Mapping[str, dc.Field]:
        return col.make_map(
            [(n, f) for f in self.fields if (n := f.metadata.get(MEMBER_NAME)) is not None],
            strict=True,
        )

    @cached.property
    def fields_by_serialization_name(self) -> ta.Mapping[str, dc.Field]:
        l = []
        for f in self.fields:
            if sn := f.metadata.get(SERIALIZATION_NAME):
                l.append((sn, f))
            elif mn := f.metadata.get(MEMBER_NAME):
                l.append((mn, f))
        return col.make_map(l, strict=True)


@dc.dataclass(frozen=True, kw_only=True)
class Shape:
    __shape__: ta.ClassVar[ShapeInfo]

    def __init_subclass__(
            cls,
            *,
            shape_name: str | None = None,
            **kwargs: ta.Any,
    ) -> None:
        super().__init_subclass__(**kwargs)

        check.state(not hasattr(cls, '__shape__'))

        info = ShapeInfo(
            cls,
            shape_metadata(**kwargs),
        )

        cls.__shape__ = info


##


class MEMBER_NAME(lang.Marker):  # noqa
    pass


class SERIALIZATION_NAME(lang.Marker):  # noqa
    pass


class VALUE_TYPE(lang.Marker):  # noqa
    pass


#


def field_metadata(
        *,
        member_name: str | None = None,
        serialization_name: str | None = None,
        value_type: ValueType | None = None,
        **kwargs: ta.Any,
) -> dict[ta.Any, ta.Any]:
    md = {**common_metadata(**kwargs)}

    if member_name is not None:
        md[MEMBER_NAME] = member_name
    if serialization_name is not None:
        md[SERIALIZATION_NAME] = serialization_name
    if value_type is not None:
        md[VALUE_TYPE] = value_type

    return md


##


@dc.dataclass(frozen=True, eq=False, kw_only=True)
class Operation:
    name: str

    input: type[Shape] | None = None
    output: type[Shape] | None = None

    errors: ta.Sequence[type[Shape]] | None = None



# ruff: noqa: UP006 UP007 UP045
from .base import Command
from .base import CommandExecutor
from .types import CommandExecutorMap


##


class LocalCommandExecutor(CommandExecutor):
    def __init__(
            self,
            *,
            command_executors: CommandExecutorMap,
    ) -> None:
        super().__init__()

        self._command_executors = command_executors

    async def execute(self, cmd: Command) -> Command.Output:
        ce: CommandExecutor = self._command_executors[type(cmd)]
        return await ce.execute(cmd)



# ruff: noqa: UP006 UP007 UP045
import dataclasses as dc

from omlish.logs.modules import get_module_logger

from ..commands.base import Command
from ..commands.base import CommandExecutor
from .deploy import DeployDriverFactory
from .specs import DeploySpec


log = get_module_logger(globals())  # noqa


##


@dc.dataclass(frozen=True)
class DeployCommand(Command['DeployCommand.Output']):
    spec: DeploySpec

    @dc.dataclass(frozen=True)
    class Output(Command.Output):
        pass


@dc.dataclass(frozen=True)
class DeployCommandExecutor(CommandExecutor[DeployCommand, DeployCommand.Output]):
    _driver_factory: DeployDriverFactory

    async def execute(self, cmd: DeployCommand) -> DeployCommand.Output:
        log.info('Deploying! %r', cmd.spec)

        with self._driver_factory(cmd.spec) as driver:
            await driver.drive_deploy()

        return DeployCommand.Output()



# ruff: noqa: UP006 UP007 UP045
import signal

from omlish.logs.modules import get_module_logger

from .groups import ProcessGroupManager
from .states import SupervisorState
from .types import ProcessOutputDispatcher
from .types import SupervisorStateManager
from .utils.signals import SignalReceiver
from .utils.signals import sig_name


log = get_module_logger(globals())  # noqa


##


class SignalHandler:
    def __init__(
            self,
            *,
            states: SupervisorStateManager,
            signal_receiver: SignalReceiver,
            process_groups: ProcessGroupManager,
    ) -> None:
        super().__init__()

        self._states = states
        self._signal_receiver = signal_receiver
        self._process_groups = process_groups

    def set_signals(self) -> None:
        self._signal_receiver.install(
            signal.SIGTERM,
            signal.SIGINT,
            signal.SIGQUIT,
            signal.SIGHUP,
            signal.SIGCHLD,
            signal.SIGUSR2,
        )

    def handle_signals(self) -> None:
        sig = self._signal_receiver.get_signal()
        if not sig:
            return

        if sig in (signal.SIGTERM, signal.SIGINT, signal.SIGQUIT):
            log.warning('received %s indicating exit request', sig_name(sig))
            self._states.set_state(SupervisorState.SHUTDOWN)

        elif sig == signal.SIGHUP:
            if self._states.state == SupervisorState.SHUTDOWN:
                log.warning('ignored %s indicating restart request (shutdown in progress)', sig_name(sig))  # noqa
            else:
                log.warning('received %s indicating restart request', sig_name(sig))  # noqa
                self._states.set_state(SupervisorState.RESTARTING)

        elif sig == signal.SIGCHLD:
            log.debug('received %s indicating a child quit', sig_name(sig))

        elif sig == signal.SIGUSR2:
            log.info('received %s indicating log reopen request', sig_name(sig))

            for p in self._process_groups.all_processes():
                for d in p.get_dispatchers():
                    if isinstance(d, ProcessOutputDispatcher):
                        d.reopen_logs()

        else:
            log.debug('received %s indicating nothing', sig_name(sig))



# ruff: noqa: TC003 UP006 UP007 UP045
import asyncio.subprocess
import dataclasses as dc
import os
import subprocess
import time
import typing as ta

from omlish.asyncs.asyncio.subprocesses import asyncio_subprocesses
from omlish.lite.check import check
from omlish.subprocesses.base import SUBPROCESS_CHANNEL_OPTION_VALUES
from omlish.subprocesses.base import SubprocessChannelOption
from omlish.subprocesses.wrap import subprocess_maybe_shell_wrap_exec

from .base import Command
from .base import CommandExecutor


##


@dc.dataclass(frozen=True)
class SubprocessCommand(Command['SubprocessCommand.Output']):
    cmd: ta.Sequence[str]

    shell: bool = False
    cwd: ta.Optional[str] = None
    env: ta.Optional[ta.Mapping[str, str]] = None

    stdout: str = 'pipe'  # SubprocessChannelOption
    decode_stdout: ta.Optional[str] = None

    stderr: str = 'pipe'  # SubprocessChannelOption
    decode_stderr: ta.Optional[str] = None

    input: ta.Optional[bytes] = None
    encode_input: ta.Optional[str] = None
    input_str: ta.Optional[str] = None

    timeout: ta.Optional[float] = None

    def __post_init__(self) -> None:
        check.not_isinstance(self.cmd, str)
        check.state(not (self.input is not None and self.input_str is not None))
        if self.decode_stdout is not None:
            check.non_empty_str(self.decode_stdout)
        if self.decode_stderr is not None:
            check.non_empty_str(self.decode_stderr)
        if self.encode_input is not None:
            check.non_empty_str(self.encode_input)

    @dc.dataclass(frozen=True)
    class Output(Command.Output):
        rc: int
        pid: int

        elapsed_s: float

        stdout: ta.Optional[bytes] = None
        stdout_str: ta.Optional[str] = None

        stderr: ta.Optional[bytes] = None
        stderr_str: ta.Optional[str] = None


class SubprocessCommandExecutor(CommandExecutor[SubprocessCommand, SubprocessCommand.Output]):
    DEFAULT_INPUT_ENCODING: ta.ClassVar[str] = 'utf-8'

    async def execute(self, cmd: SubprocessCommand) -> SubprocessCommand.Output:
        input_bytes: ta.Optional[bytes]
        if cmd.input is not None:
            check.none(cmd.input_str)
            input_bytes = cmd.input
        elif cmd.input_str is not None:
            input_bytes = cmd.input_str.encode(cmd.encode_input or self.DEFAULT_INPUT_ENCODING)
        else:
            input_bytes = None

        proc: asyncio.subprocess.Process
        async with asyncio_subprocesses.popen(
            *subprocess_maybe_shell_wrap_exec(*cmd.cmd),

            shell=cmd.shell,
            cwd=cmd.cwd,
            env={**os.environ, **(cmd.env or {})},

            stdin=subprocess.PIPE if input_bytes is not None else None,
            stdout=SUBPROCESS_CHANNEL_OPTION_VALUES[ta.cast(SubprocessChannelOption, cmd.stdout)],
            stderr=SUBPROCESS_CHANNEL_OPTION_VALUES[ta.cast(SubprocessChannelOption, cmd.stderr)],

            timeout=cmd.timeout,
        ) as proc:
            start_time = time.time()
            stdout, stderr = await asyncio_subprocesses.communicate(
                proc,
                input=input,
                timeout=cmd.timeout,
            )
            end_time = time.time()

        out_kw: ta.Dict[str, ta.Any] = {}
        if stdout is not None:
            if cmd.decode_stdout is not None:
                out_kw.update(stdout_str=stdout.decode(cmd.decode_stdout))
            else:
                out_kw.update(stdout=stdout)
        if stderr is not None:
            if cmd.decode_stderr is not None:
                out_kw.update(stderr_str=stderr.decode(cmd.decode_stderr))
            else:
                out_kw.update(stderr=stderr)

        return SubprocessCommand.Output(
            rc=check.not_none(proc.returncode),
            pid=proc.pid,

            elapsed_s=end_time - start_time,

            **out_kw,
        )



import dataclasses as dc
import sys

from omlish.lite.abstract import Abstract
from omlish.lite.cached import cached_nullary
from omlish.logs.modules import get_module_logger
from omlish.os.linux import LinuxOsRelease


log = get_module_logger(globals())  # noqa


##


@dc.dataclass(frozen=True)
class Platform(Abstract):
    pass


class LinuxPlatform(Platform, Abstract):
    pass


class UbuntuPlatform(LinuxPlatform):
    pass


class AmazonLinuxPlatform(LinuxPlatform):
    pass


class GenericLinuxPlatform(LinuxPlatform):
    pass


class DarwinPlatform(Platform):
    pass


class UnknownPlatform(Platform):
    pass


##


def _detect_system_platform() -> Platform:
    plat = sys.platform

    if plat == 'linux':
        if (osr := LinuxOsRelease.read()) is None:
            return GenericLinuxPlatform()

        if osr.id == 'amzn':
            return AmazonLinuxPlatform()

        elif osr.id == 'ubuntu':
            return UbuntuPlatform()

        else:
            return GenericLinuxPlatform()

    elif plat == 'darwin':
        return DarwinPlatform()

    else:
        return UnknownPlatform()


@cached_nullary
def detect_system_platform() -> Platform:
    platform = _detect_system_platform()
    log.info('Detected platform: %r', platform)
    return platform



# ruff: noqa: UP006 UP007 UP045
"""
TODO:
 - sequence all messages
"""
import asyncio
import dataclasses as dc
import itertools
import logging
import time
import typing as ta

from omlish.lite.abstract import Abstract
from omlish.lite.check import check
from omlish.logs.modules import get_module_logger

from ..commands.base import Command
from ..commands.base import CommandException
from ..commands.base import CommandExecutor
from ..commands.base import CommandOutputOrException
from ..commands.base import CommandOutputOrExceptionData
from .channel import RemoteChannel


T = ta.TypeVar('T')


log = get_module_logger(globals())  # noqa


##


class _RemoteProtocol:
    class Message(Abstract):
        async def send(self, chan: RemoteChannel) -> None:
            await chan.send_obj(self, _RemoteProtocol.Message)

        @classmethod
        async def recv(cls: ta.Type[T], chan: RemoteChannel) -> ta.Optional[T]:
            return await chan.recv_obj(cls)

    #

    class Request(Message, Abstract):
        pass

    @dc.dataclass(frozen=True)
    class CommandRequest(Request):
        seq: int
        cmd: Command

    @dc.dataclass(frozen=True)
    class PingRequest(Request):
        time: float

    #

    class Response(Message, Abstract):
        pass

    @dc.dataclass(frozen=True)
    class LogResponse(Response):
        s: str

    @dc.dataclass(frozen=True)
    class CommandResponse(Response):
        seq: int
        res: CommandOutputOrExceptionData

    @dc.dataclass(frozen=True)
    class PingResponse(Response):
        time: float


##


class _RemoteLogHandler(logging.Handler):
    def __init__(
            self,
            chan: RemoteChannel,
            loop: ta.Any = None,
    ) -> None:
        super().__init__()

        self._chan = chan
        self._loop = loop

    def emit(self, record):
        msg = self.format(record)

        async def inner():
            await _RemoteProtocol.LogResponse(msg).send(self._chan)

        loop = self._loop
        if loop is None:
            loop = asyncio.get_running_loop()
        if loop is not None:
            asyncio.run_coroutine_threadsafe(inner(), loop)


##


class _RemoteCommandHandler:
    DEFAULT_PING_INTERVAL_S: float = 3.

    def __init__(
            self,
            chan: RemoteChannel,
            executor: CommandExecutor,
            *,
            stop: ta.Optional[asyncio.Event] = None,
            ping_interval_s: float = DEFAULT_PING_INTERVAL_S,
    ) -> None:
        super().__init__()

        self._chan = chan
        self._executor = executor
        self._stop = stop if stop is not None else asyncio.Event()
        self._ping_interval_s = ping_interval_s

        self._cmds_by_seq: ta.Dict[int, _RemoteCommandHandler._Command] = {}

        self._last_ping_send: float = 0.
        self._ping_in_flight: bool = False
        self._last_ping_recv: ta.Optional[float] = None

    def stop(self) -> None:
        self._stop.set()

    @dc.dataclass(frozen=True)
    class _Command:
        req: _RemoteProtocol.CommandRequest
        fut: asyncio.Future

    async def run(self) -> None:
        log.debug('_RemoteCommandHandler loop start: %r', self)

        stop_task = asyncio.create_task(self._stop.wait())
        recv_task: ta.Optional[asyncio.Task] = None

        while not self._stop.is_set():
            if recv_task is None:
                recv_task = asyncio.create_task(_RemoteProtocol.Message.recv(self._chan))

            if not self._ping_in_flight:
                if not self._last_ping_recv:
                    ping_wait_time = 0.
                else:
                    ping_wait_time = self._ping_interval_s - (time.time() - self._last_ping_recv)
            else:
                ping_wait_time = float('inf')
            wait_time = min(self._ping_interval_s, ping_wait_time)
            log.debug('_RemoteCommandHandler loop wait: %f', wait_time)

            done, pending = await asyncio.wait(
                [
                    stop_task,
                    recv_task,
                ],
                return_when=asyncio.FIRST_COMPLETED,
                timeout=wait_time,
            )

            #

            if (
                    (time.time() - self._last_ping_send >= self._ping_interval_s) and
                    not self._ping_in_flight
            ):
                now = time.time()
                self._last_ping_send = now
                self._ping_in_flight = True
                await _RemoteProtocol.PingRequest(
                    time=now,
                ).send(self._chan)

            #

            if recv_task in done:
                msg: ta.Optional[_RemoteProtocol.Message] = check.isinstance(
                    recv_task.result(),
                    (_RemoteProtocol.Message, type(None)),
                )
                recv_task = None

                if msg is None:
                    break

                await self._handle_message(msg)

        log.debug('_RemoteCommandHandler loop stopping: %r', self)

        for task in [
            stop_task,
            recv_task,
        ]:
            if task is not None and not task.done():
                task.cancel()

        for cmd in self._cmds_by_seq.values():
            cmd.fut.cancel()

        log.debug('_RemoteCommandHandler loop exited: %r', self)

    async def _handle_message(self, msg: _RemoteProtocol.Message) -> None:
        if isinstance(msg, _RemoteProtocol.PingRequest):
            log.debug('Ping: %r', msg)
            await _RemoteProtocol.PingResponse(
                time=msg.time,
            ).send(self._chan)

        elif isinstance(msg, _RemoteProtocol.PingResponse):
            latency_s = time.time() - msg.time
            log.debug('Pong: %0.2f ms %r', latency_s * 1000., msg)
            self._last_ping_recv = time.time()
            self._ping_in_flight = False

        elif isinstance(msg, _RemoteProtocol.CommandRequest):
            fut = asyncio.create_task(self._handle_command_request(msg))
            self._cmds_by_seq[msg.seq] = _RemoteCommandHandler._Command(
                req=msg,
                fut=fut,
            )

        else:
            raise TypeError(msg)

    async def _handle_command_request(self, req: _RemoteProtocol.CommandRequest) -> None:
        res = await self._executor.try_execute(
            req.cmd,
            log=log,
            omit_exc_object=True,
        )

        await _RemoteProtocol.CommandResponse(
            seq=req.seq,
            res=CommandOutputOrExceptionData(
                output=res.output,
                exception=res.exception,
            ),
        ).send(self._chan)

        self._cmds_by_seq.pop(req.seq)  # noqa


##


@dc.dataclass()
class RemoteCommandError(Exception):
    e: CommandException


class RemoteCommandExecutor(CommandExecutor):
    def __init__(self, chan: RemoteChannel) -> None:
        super().__init__()

        self._chan = chan

        self._cmd_seq = itertools.count()
        self._queue: asyncio.Queue = asyncio.Queue()  # asyncio.Queue[RemoteCommandExecutor._Request]
        self._stop = asyncio.Event()
        self._loop_task: ta.Optional[asyncio.Task] = None
        self._reqs_by_seq: ta.Dict[int, RemoteCommandExecutor._Request] = {}

    #

    async def start(self) -> None:
        check.none(self._loop_task)
        check.state(not self._stop.is_set())
        self._loop_task = asyncio.create_task(self._loop())

    async def aclose(self) -> None:
        self._stop.set()
        if self._loop_task is not None:
            await self._loop_task

    #

    @dc.dataclass(frozen=True)
    class _Request:
        seq: int
        cmd: Command
        fut: asyncio.Future

    async def _loop(self) -> None:
        log.debug('RemoteCommandExecutor loop start: %r', self)

        stop_task = asyncio.create_task(self._stop.wait())
        queue_task: ta.Optional[asyncio.Task] = None
        recv_task: ta.Optional[asyncio.Task] = None

        while not self._stop.is_set():
            if queue_task is None:
                queue_task = asyncio.create_task(self._queue.get())
            if recv_task is None:
                recv_task = asyncio.create_task(_RemoteProtocol.Message.recv(self._chan))

            done, pending = await asyncio.wait(
                [
                    stop_task,
                    queue_task,
                    recv_task,
                ],
                return_when=asyncio.FIRST_COMPLETED,
            )

            #

            if queue_task in done:
                req = check.isinstance(queue_task.result(), RemoteCommandExecutor._Request)
                queue_task = None
                await self._handle_queued_request(req)

            #

            if recv_task in done:
                msg: ta.Optional[_RemoteProtocol.Message] = check.isinstance(
                    recv_task.result(),
                    (_RemoteProtocol.Message, type(None)),
                )
                recv_task = None

                if msg is None:
                    log.debug('RemoteCommandExecutor got eof: %r', self)
                    break

                await self._handle_message(msg)

        log.debug('RemoteCommandExecutor loop stopping: %r', self)

        for task in [
            stop_task,
            queue_task,
            recv_task,
        ]:
            if task is not None and not task.done():
                task.cancel()

        for req in self._reqs_by_seq.values():
            req.fut.cancel()

        log.debug('RemoteCommandExecutor loop exited: %r', self)

    async def _handle_queued_request(self, req: _Request) -> None:
        self._reqs_by_seq[req.seq] = req
        await _RemoteProtocol.CommandRequest(
            seq=req.seq,
            cmd=req.cmd,
        ).send(self._chan)

    async def _handle_message(self, msg: _RemoteProtocol.Message) -> None:
        if isinstance(msg, _RemoteProtocol.PingRequest):
            log.debug('Ping: %r', msg)
            await _RemoteProtocol.PingResponse(
                time=msg.time,
            ).send(self._chan)

        elif isinstance(msg, _RemoteProtocol.PingResponse):
            latency_s = time.time() - msg.time
            log.debug('Pong: %0.2f ms %r', latency_s * 1000., msg)

        elif isinstance(msg, _RemoteProtocol.LogResponse):
            log.info(msg.s)

        elif isinstance(msg, _RemoteProtocol.CommandResponse):
            req = self._reqs_by_seq.pop(msg.seq)
            req.fut.set_result(msg.res)

        else:
            raise TypeError(msg)

    #

    async def _remote_execute(self, cmd: Command) -> CommandOutputOrException:
        req = RemoteCommandExecutor._Request(
            seq=next(self._cmd_seq),
            cmd=cmd,
            fut=asyncio.Future(),
        )
        await self._queue.put(req)
        return await req.fut

    # @ta.override
    async def execute(self, cmd: Command) -> Command.Output:
        r = await self._remote_execute(cmd)
        if (e := r.exception) is not None:
            raise RemoteCommandError(e)
        else:
            return check.not_none(r.output)

    # @ta.override
    async def try_execute(
            self,
            cmd: Command,
            *,
            log: ta.Optional[logging.Logger] = None,  # noqa
            omit_exc_object: bool = False,
    ) -> CommandOutputOrException:
        try:
            r = await self._remote_execute(cmd)

        except Exception as e:  # noqa
            if log is not None:
                log.exception('Exception executing remote command: %r', type(cmd))

            return CommandOutputOrExceptionData(exception=CommandException.of(
                e,
                omit_exc_object=omit_exc_object,
                cmd=cmd,
            ))

        else:
            return r
